package jssc_test;
import jssc.SerialPortList;
import jssc.SerialPort;
import jssc.SerialPortException;
import java.util.*;
import java.io.*;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import jssc.SerialPortEvent;
import jssc.SerialPortEventListener;

public class JSSC_test {
    
    //Serial Communication Variables
    public static SerialPort serialPort;
    static int mask;
    
    //Setup the flag variable to prevent too much gcode dripping
    public static String inputStr = "ok";
    
    //Setup menu variable
    public static String menu = "menu";

    /*/////////////////////////////////////////////////////////////////////////    
    //Calibration Constants - Edit these for any change in tape and reels
    //*************************************************************************

    Reels
        1 = .................nothing
        2 = 400k ohm
        3 = 10k ohm
        4 = 75k ohm
        5 = 100k ohm
        6 = 51k ohm
        7 = 25k ohm
        8 = 5k ohm
        9 = 22 pF
        10 = 100 nF
        11 = 33 pF
        12 = ................nothing
        13 = .01 uF
        14 = 2.5 kohm
        15 = 200 kohm
        16 = 300 ohm
        17 = ................nothing
    
    Trays
        1 = lm324
        2 = zener
        3 = .47uF
        4 = 1uF
    
    
    Reel Tape X and Y - are the absolute positions from origin where the nozzle
    will go to induce a tape retraction. In other words, from the standard view
    the machine will move to the tape, lower z axis, and move to the left 4 mm.
    Change these values as you swap out the feeder guides.
    
    Reel Posit X and Y - are the RELATIVE positions from Reel Tape X and Y after
    before a retraction. This means if tapex and tapey were (100,100) before 
    retraction, the retraction will occur and the nozzle will move to
    (100 - reelPositX, 100 - reelPositY). Change these values if alignment is 
    off or you swap new components.
    
    Reel Height - is the z height above origin which the nozzle will touch the
    component on the tape. Adjust this accordingly to allow a little contact
    with the component. Crashing will not occur if you built the machine as I
    did. The weight of the zaxis guide rod, nozzle, and a little bit of spring
    force will not hurt your components.
    
    Steps to Advance - is the number of times you need pull the tape in 4mm 
    increments to expose a full new components. Adjust this for anything bigger
    than a 0805 tape (ex. soic8 has components centered at every 8mm, so two
    4mm  retraction cycles need to occur for machine to expose another component
    for placing
  
    */  
    
    public static final int xyspeed = 6500;             //max table speed in mm/min
    public static final int zspeed =  300;              //max z axis speed in mm/min
    public static final int maxx =    240;              //max positive x position in mm from origin
    public static final int maxy =    240;              //max positive y position in mm from origin
    public static final int maxz =    120;              //max positive z position in mm from origin
    public static final double homeX = 1.10;            //home x position after auto home
    public static final double homeY = 3.00;            //home y position after auto home
    public static final double homeZ = 9.32;           //home z after max limit reached
    public static final boolean invertNozzle = true;    //invert nozzle operation
    public static final double standardSafetyHeight = 7.5;     //standard retraction height in mm
    public static final int retractionSpeed = 150;      // retraction speed in mm/min
    public static final double retractionX = -4.60;     //pull the tape mm out before picking up component
    public static final double backRetractionX = 0.80;  //push the tape mm in after retraction 
    public static final double slopRetractionX = -0.35; //center in middle of hole of tape after retract and backretract
    public static final double retractionDropHeight = 2.75; //height to drop the needle for retraction
    public static final double boardHeight = 1.10;       //target board height in seconds
    public static final int suctionDwell = 100;         //suction dwell period in seconds
    public static final int releaseDwell = 100;         //release dwell period in seconds
    public static final String fileDirectory = "run";
    public static final String configDirectory = "config";
    
    public static final int reelTotal =  17;     //   1       2       3       4       5       6       7       8       9      10      11      12      13      14      15      16      17
    public static final double reelTapeX[] =  { 238.00, 238.00, 238.00, 238.00, 236.00, 236.00, 236.00, 236.00, 236.00, 236.00, 236.00, 236.00, 236.00, 236.00, 236.00, 236.00, 236.00};
    public static final double reelTapeY[] =  { 208.10, 197.00, 186.00, 175.00, 164.00, 153.10, 142.10, 131.10, 120.10, 109.10,  98.10,  87.10,  76.10,  65.10,  54.10,  43.10,  32.10};
    public static final double reelPositX[] = {      6,      6,      6,      6,      6,      6,      6,      6,      6,      6,      6,      6,      6,      6,      6,      6,      6};
    public static final double reelPositY[] = {    3.5,    3.5,    3.5,    3.5,    3.5,    3.5,    3.5,    3.5,    3.5,    3.5,    3.5,    3.5,    3.5,    3.5,    3.5,    3.5,    3.5};
    public static final double reelHeight[] = {   3.15,   3.15,   3.15,   3.15,   3.15,   3.15,   3.15,   3.15,   3.15,   3.15,   3.15,   3.15,   3.15,   3.15,   3.15,   3.15,   3.15};
    public static final int stepsToAdvance[] ={      1,      1,      1,      1,      1,      1,      1,      1,      1,      1,      1,      1,      1,      1,      1,      1,      1};

    //trays
    // you need to program each position of each component in a tray.
    // added note: these positions are based on home, subtraction added for home offset in runFile()
    //trays
    public static final int trayTotal = 4;
    public static final int trayCount[] = {
        21, //total number of components in tray 1
        21, //"                                " 2
        21, //"                                " 3
        21  //"                                " 4
    };
    public static final double trayX[][] = {
    //posit   1      2      3      4      5      6      7      8      9     10     11     12     13     14     15     16     17     18     19     20     21     
        { 20.25, 28.25, 36.25, 44.25, 52.25, 60.25, 68.25, 76.25, 84.25, 92.25,100.25,108.25,116.25,124.25,132.25,140.25,148.25,156.25,164.25,172.25,180.25},  //tray 1
        { 19.75, 23.75, 27.75, 31.75, 35.75, 39.75, 43.75, 47.75, 51.75, 55.75, 60.75, 64.75, 68.75, 72.75, 76.75, 80.75, 84.75, 88.75, 92.75, 96.75,100.75},  //tray 2
        { 19.75, 23.75, 27.75, 31.75, 35.75, 39.75, 43.75, 47.75, 51.75, 55.75, 60.75, 64.75, 68.75, 72.75, 76.75, 80.75, 84.75, 88.75, 92.75, 96.75,100.75},  //tray 3
        { 19.75, 23.75, 27.75, 31.75, 35.75, 39.75, 43.75, 47.75, 51.75, 55.75, 60.75, 64.75, 68.75, 72.75, 76.75, 80.75, 84.75, 88.75, 92.75, 96.75,100.75}   //tray 4
    };
    public static final double trayY[][] = {
    //posit   1      2      3      4      5      6      7      8      9     10     11     12     13     14     15     16     17     18     19     20     21
        {169.85,169.85,169.85,169.85,169.85,169.85,169.85,169.85,169.85,169.85,169.85,169.85,169.85,169.85,169.85,169.85,169.85,169.85,169.85,169.85,169.85},  //tray 1
        {211.85,211.85,211.85,211.85,211.85,211.85,211.85,211.85,211.85,211.85,211.85,211.85,211.85,211.85,211.85,211.85,211.85,211.85,211.85,211.85,211.85},  //tray 2
        {224.65,224.65,224.65,224.65,224.65,224.65,224.65,224.65,224.65,224.65,224.65,224.65,224.65,224.65,224.65,224.65,224.65,224.65,224.65,224.65,224.65},  //tray 3
        {237.65,237.65,237.65,237.65,237.65,237.65,237.65,237.65,237.65,237.65,237.65,237.65,237.65,237.65,237.65,237.65,237.65,237.65,237.65,237.65,237.65}  //tray 4
    };
    public static final double trayZ[] = {
    //tray    1      2      3      4 
           1.00,  1.00,  1.00,  1.00
    };
    
    
    //*************************************************************************
    ///////////////////////////////////////////////////////////////////////////
    
    //Main loop
    public static void main(String[] args) throws FileNotFoundException, IOException, SerialPortException {
        
        //Intro header to user
        Scanner console = new Scanner(System.in);
        System.out.println("<--- WELCOME TO PICK and PLACE COMMUNICATOR --->\n");
        
        //print name of all available ports to communicate with
        printPorts();
        
        //select port to open
        selectPortToOpen(console);
        
        //startup routine
        pnpStartUpRoutine();
        
        //Main menu loops until exit is received
        while(true){
            
            //Get what user wants to do
            inputSelection(console);
            
            //calibration check to perform slow table travel, fast table travel, then reelcheck
            if(menu.equals("calib")){
                System.out.println("Ensure you are zeroed and the table is clear, then press enter");
                System.in.read();
                tableExtents("slow", serialPort);
                System.out.println("Slow run complete, press enter to perform fast speed check");
                System.in.read();
                tableExtents("fast", serialPort);
                System.out.println("Fast run complete");
                System.out.println("Table Extents calibraiton complete\n\n");
            }
            
            //Perform Pick and Place of Programmed File
            if(menu.equals("run")){
                runFile(console);
            }
            
            //Manual Input Commands
            if(menu.equals("manual")){
                manualInput(console);
            }
            
            //alignment of tape and reels
            if(menu.equals("alignTape")){
                System.out.println("Performing Reel Check");
                reelCheck(console);
                System.out.println("Reel Check Complete");
            }
            
            //run test of file
            if(menu.equals("test")){
                System.out.println("Performing test of file");
                runFileTest(console);
            }
            
            //exit the program
            if(menu.equals("exit")){
                System.out.println("\n\nEnding Process. Goodbye");
                System.exit(0);
            }
        }
    }
    
    public static void pnpStartUpRoutine() throws SerialPortException{
        
        //Wait for PNP to spool up
        System.out.println("Machine Startup Routine Running. Please Wait\n");
        try{
            Thread.sleep(6000);
        }
        catch(Exception ex){
            System.out.println(ex);
        }
        
        //machine setup gcode commands
        pnpWrite("G21");  //set units to millimeters
        pnpWrite("G90");  //start in absolute coordinate system
        String GCode = "G1 Z20 F" + Integer.toString(zspeed); //raise z axis to limit
        pnpWrite(GCode);
        pnpWrite("M280 P0 S0");//turn nozzle to zero degrees
        GCode = "G92 Z" + Double.toString(homeZ); //set z axis to offset position
        pnpWrite(GCode);
        pnpWrite("G28 X Y"); //home x and y
        GCode = "G1 X" + Double.toString(homeX) + " Y" + Double.toString(homeY) + " F300"; //move to offset home
        pnpWrite(GCode);
        pnpWrite("G92 X0 Y0"); //tell machine it is at 0,0
        GCode = "G1 Z0.00 F" + Integer.toString(zspeed); //drop nozzle to zero
        pnpWrite(GCode);
        System.out.println("PNP GCode Setup Commands Sent\nMachine Ready");
    }
    
    //Printout the available ports
    public static void printPorts(){
        String[] portNames = SerialPortList.getPortNames();
        System.out.println("AVAILABLE PORTS:");
        for(int i = 0; i < portNames.length; i++){
            System.out.println(i+1 + ".) "+ portNames[i]);
        }
        System.out.println();
    }
    
    //Ask for port to open, and open it, set parameters for communicating
    public static void selectPortToOpen(Scanner console) throws SerialPortException{
        
        //Which Port to  open?
        System.out.print("ENTER PORT TO OPEN: " );
        String arduinoPort = console.next();
        System.out.println("You have chosen \"" + arduinoPort + "\" to communicate with\n");
        serialPort = new SerialPort(arduinoPort);
        
        //would you like to open port?
        System.out.print("OPEN PORT? (yes or no): ");
        String inputMode = console.next();
        System.out.println();
        
        //Open Port, set parameters
        if(inputMode.equals("yes")){
            try {
                System.out.println("Port opened: " + serialPort.openPort());
                //Set port paramters: baud, data bit length, stop bit, parity, setRTS, set DTR 
                System.out.println("Parameters set: " + serialPort.setParams(115200, 8, 1, 0, false,true));
                int mask = SerialPort.MASK_RXCHAR + SerialPort.MASK_CTS + SerialPort.MASK_DSR;//Prepare mask
                serialPort.setEventsMask(mask);//Set mask
                serialPort.addEventListener(new SerialPortReader()); //add event listener
            }
            catch (SerialPortException ex){
                System.out.println(ex);
            }
        }
    }
    
    //Select type of input
    public static void inputSelection(Scanner console) throws FileNotFoundException, IOException{
        int menunumber = 0;
        while(menunumber > 5 || menunumber < 1){
            System.out.print("\nWHAT WOULD YOU LIKE TO DO?"
                           + "\n 1. Calibrate Speed"
                           + "\n 2. Run File"
                           + "\n 3. Manual Input"
                           + "\n 4. Tape and Reel Alignment"
                           + "\n 5. Test File"
                           + "\n 6. Exit"
                           + "\n Enter Selection: ");
            menunumber = console.nextInt();
            System.out.println();
        }
        if(menunumber == 1){
            menu = "calib";
        }
        else
        if(menunumber == 2){
            menu = "run";
        }
        else
        if(menunumber == 3){
            menu = "manual";
        }else
        if(menunumber == 4){
            menu = "alignTape";
        }else
        if(menunumber == 5){
            menu = "test";
        }else
        if(menunumber == 6){
            menu = "exit";
        }
    }
    
    //Manual input of GCode to pnp
    public static void manualInput(Scanner console) throws IOException, SerialPortException{
        inputStr = "ok";
        //Introduction of how to put in coordinates
        System.out.println("Type in GCode to send to Pick and Place");
        System.out.println("type 'no' to exit at end of command. Example below\n");
        //Input Manual GCode
        do{
            System.out.print("Input G Code: ");
            BufferedReader br = new BufferedReader(new InputStreamReader(System.in));
            String GCode = br.readLine();
            System.out.println(GCode);
            System.out.print("Is this correct? tye 'no' if not:  ");
            String evalString = console.next();
            if(evalString.equals("yes")){
                break;
            }
            System.out.println("Writing to Port\nPNP busy\n");
            pnpWrite(GCode);
        }
        while(true);
    }
    
    //go to table extents
    public static void tableExtents(String speed, SerialPort serialPort) throws SerialPortException{
        
        //adjust moveSpeed variable appropriately
        int moveSpeed = 0;
        int repeat = 1;
        if(speed.equals("slow")){
            moveSpeed = 2000; //millimeters per second
            repeat = 1;
        }else
        if(speed.equals("fast")){
            moveSpeed = xyspeed;
            repeat = 3;
        }
        
        //perform travel moves, more if doing fast speed test
        for(int i =0; i < repeat; i++){
            //Initilize variable
            String GCode = "";
            
            //lift nozzle to standard height to prevent crash
            GCode = "G1 Z" + Double.toString(standardSafetyHeight) + " F" + Integer.toString(zspeed);
            pnpWrite(GCode);

            //move to top left
            GCode = "G1 Y" + Integer.toString(maxy) + " F" + Integer.toString(moveSpeed);
            pnpWrite(GCode);

            //move to top right
            GCode = "G1 X" + Integer.toString(maxx) + " F" + Integer.toString(moveSpeed);
            pnpWrite(GCode);

            //move to bottom right
            GCode = "G1 Y0 F" + Integer.toString(moveSpeed);
            pnpWrite(GCode);

            //move to bottom left
            GCode = "G1 X0 F" + Integer.toString(moveSpeed);
            pnpWrite(GCode);

            //drop nozzle to 0
            GCode = "G1 Z0 F" + Integer.toString(zspeed);
            pnpWrite(GCode);
        }
    }
    
    //check positions of reels
    public static void reelCheck(Scanner console) throws SerialPortException, IOException{
    
        //initialize variable
        String GCode = "";

        //lift nozzle to prevent crash
        pnpWrite("G1 Z" + Double.toString(standardSafetyHeight) + " F" + Integer.toString(zspeed));
        
        //turn off suction
        if(invertNozzle == true){
            pnpWrite("M106");
        }else{
            pnpWrite("M107");
        }
        
        //go to each starting retraction point
        for(int i =0; i< reelTotal; i++){
            
            //go in front of reel before position
            GCode = "G1 X" + Double.toString(reelTapeX[i] - 30.00) + " Y" + Double.toString(reelTapeY[i]) + " F" + Integer.toString(xyspeed);
            pnpWrite(GCode);
            
            //go to position
            GCode = "G1 X" + Double.toString(reelTapeX[i]) + " Y" + Double.toString(reelTapeY[i]) + " F" + Integer.toString(xyspeed);
            pnpWrite(GCode);
            
            //drop nozzle for retraction verification
            double drop = retractionDropHeight;
            GCode = "G1 Z" + Double.toString(drop) + " F" + Integer.toString(zspeed);
            pnpWrite(GCode);
            
            //ask to advance
            System.out.println("Press enter to continue");
            System.in.read();
            
            //raise nozzle
            GCode = "G1 Z" + Double.toString(standardSafetyHeight) + " F" + Integer.toString(zspeed);
            pnpWrite(GCode);
            
            //center on current component
//            double positx = reelTapeX[i] - reelPositX[i];
//            double posity = reelTapeY[i] + reelPositY[i];
//            GCode = "G1 X" + Double.toString(positx) + " Y" + Double.toString(posity) + " F" + Integer.toString(xyspeed);
//            pnpWrite(GCode);
            
            //drop nozzle to component height to check alignment
//            GCode = "G1 Z" + Double.toString(reelHeight[i]) + " F" + Integer.toString(zspeed);
//            pnpWrite(GCode);
            
            //ask to advance
//            System.out.println("Press enter to continue");
//            System.in.read();
            
            //raise nozzle
//            GCode = "G1 Z" + Double.toString(standardSafetyHeight) + " F" + Integer.toString(zspeed);
//            pnpWrite(GCode);
        }
        
        //return to origin
        pnpWrite("G1 X0 Y0 F" + Integer.toString(xyspeed));
        pnpWrite("G1 Z0 F" + Integer.toString(zspeed));
    }
    
    //Run File execution
    public static void runFile(Scanner console) throws FileNotFoundException, SerialPortException, IOException{
        
        //ensure at origin
        pnpWrite("G1 X0 Y0 Z0 F1000");
        
        //turn off nozzle suction
        if(invertNozzle == true){
            pnpWrite("M106");
        }else{
            pnpWrite("M107");
        }
        
        //search directory and list names of files
        File folder = new File(fileDirectory);
        File[] listOfFiles = folder.listFiles();
        System.out.println("\nContents of Directory\n");
        for(int i = 0; i < listOfFiles.length; i++) {
            if (listOfFiles[i].isFile()) {
                System.out.println("File:    " + listOfFiles[i].getName());
            } else if (listOfFiles[i].isDirectory()) {
                System.out.println("Folder:  " + listOfFiles[i].getName());
            }
        }
        
        //ask which file to run
        System.out.print("Which File to run? include extension: ");
        String fileName = console.next();
        String fullpath = fileDirectory + "/" + fileName;
        File GCodeFile = new File(fullpath);
        
        //Check if file exists
        while(!GCodeFile.exists()){
            System.out.println("File does not exist!!! ");
            System.out.print("Enter file name with extension: ");
            fileName = console.next();
            GCodeFile = new File(fileName);
        }
        System.out.println("File " + fileName + " opened.");
        
        //Create a follow up file that records all compeleted operations
        PrintStream backUpFile = new PrintStream("currentProject.txt");
        System.out.println("currentProject.txt was created \n");
        
        //ask to commence execution
        System.out.print("Commence Execution? (yes or no) ");
        fileName = console.next();
        if(fileName.equals("no"))
            return;
        
        /*
        Precoded file structure is as follows:
        
        First Token # - set new absolute x and y coordinates
        First Token "R" or "T" - reel or tape position. reel automatically assumes
           retraction cycle and tray relies on manual coded positions 
        Second Token - component from reel 1 will be retracted and picked up
        Third Token - x coordinate of where component is centered to go
        Fourth Token - y coordinate of where component is centered to go
        Fifth Token - Rotation of object prior to putting on board
        
        Example: "# 40 50"
        Example: "R 1 X95.00 Y90.00 S0"
        
        Without the quotes, the cnc will set all relative coordinates of
        components to place based on x of 40 and y of 50 mm. If nothing set, the
        default is origin of 0,0. Next the cnc will go to first reel ("R"), 
        retract component, pick it up from the reel, move to (95,90), rotate to 
        0 degrees, and place the component
        
        Example: "T 7 X95.00 Y92.00 S80"
        
        cnc will go to 7th tray and automatically seek the next component of
        that tray, pick it up, move to (95,92), rotate 80 degrees, and place
        component
        */
        
        //setup common variables
        String discard =    "0";
        String x_absolute = "0";
        String y_absolute = "0";
        double xabs = 0.00;
        double yabs = 0.00; 
        String GCode = "0";
        double drop = 0.00;
        boolean setAbsolute = false;
        
        //execute file
        Scanner gcodeScanner = new Scanner(GCodeFile);
        while(gcodeScanner.hasNextLine()){
         
            //reset the set absolute attribute
            setAbsolute = false;
            
            //Consume line, copy it to backup file in case of mid pnp failure
            String incomingString = gcodeScanner.nextLine();
            backUpFile.println(incomingString);
            Scanner incomingToken = new Scanner(incomingString);
            
            //check if new relative position expected
            char checkChar = incomingString.charAt(0);
            if(checkChar == '#'){
                setAbsolute = true;
            }
            
            //set absolute coordinates if token received to do so (received #)
            if(setAbsolute == true){
                
                //expect three tokens
                discard =   incomingToken.next();  //the # symbol needs to be consumed by scanner
                x_absolute = incomingToken.next();
                y_absolute = incomingToken.next();
                
                //define new absolute coordinates
                xabs = Double.parseDouble(x_absolute);
                yabs = Double.parseDouble(y_absolute);
                
                //tell user new absolute coordinates set
                System.out.println("\nNew Board Coordinates Set");
                System.out.println("X Posit: " + xabs);
                System.out.println("Y Posit: " + yabs + "\n");

                //lift head and go to new board coordinates
                GCode = "G1 Z" + Double.toString(standardSafetyHeight) + " F" + Integer.toString(zspeed);
                pnpWrite(GCode);
                GCode = "G1 X" + Double.toString(xabs) + " Y" + Double.toString(yabs) + " F" + Integer.toString(xyspeed);
                pnpWrite(GCode);
                
                //lower to board height
                drop = boardHeight;
                GCode = "G1 Z" + Double.toString(drop) + " F" + Integer.toString(zspeed);
                pnpWrite(GCode);
                
                //ask for permission to continue
                System.out.println("Press enter to Move head out of the way");
                System.in.read();
                
                //lift head and go to origin to get out of way if rotation required
                GCode = "G1 Z" + Double.toString(standardSafetyHeight) + " F" + Integer.toString(zspeed);
                pnpWrite(GCode);
                pnpWrite("G1 X0 Y0 F" + Integer.toString(xyspeed));
                
                //ask for permission to continue
                System.out.println("Press enter to Continue Pick and Place");
                System.in.read();
            }
            
            //if no new absolute coordinate, execute the pick and place of components listed
            else {
                //5 tokens are expected, put them in variables for manipulation
                String feedType = incomingToken.next();    //feed type either "R" or "T" for reel or tray
                String component = incomingToken.next();   //component number of feedType
                String xposit = incomingToken.next();      //x position to place component
                String yposit = incomingToken.next();      //y position to place component
                String rotationDeg = incomingToken.next(); //rotation of component
                
                //turn tokens into other variables as necessary
                char feedVerify = feedType.charAt(0);         
                int pickUp = Integer.parseInt(component) - 1;
                double parseX = Double.parseDouble(xposit) + xabs; //determine absolute coordinate
                double parseY = Double.parseDouble(yposit) + yabs; //determine absolute coordinate
                xposit = Double.toString(parseX); //absolute x coordinate of component to move to
                yposit = Double.toString(parseY); //absolute y coordinate of component to move to

                //lift nozzle to standard height
                GCode = "G1 Z" + Double.toString(standardSafetyHeight) + " F" + Integer.toString(zspeed);
                pnpWrite(GCode);
                
                //check to make sure nozzle is at 0 degrees
                pnpWrite("G4 P1");
                pnpWrite("M280 P0 S0");

                //if component is on reel, go retract and place head over it
                if(feedVerify == 'R'){
                    
                    //Retraction Procedure
                    for(int i = 0; i<stepsToAdvance[pickUp]; i++){

                        //move 3cm infront of tape position to prevent scraping
                        GCode = "G1 X" + Double.toString(reelTapeX[pickUp] - 30.00) + " Y" + Double.toString(reelTapeY[pickUp]) + " F" + Integer.toString(xyspeed);
                        pnpWrite(GCode);
                        
                        //move to tape position for retraction
                        GCode = "G1 X" + Double.toString(reelTapeX[pickUp]) + " Y" + Double.toString(reelTapeY[pickUp]) + " F" + Integer.toString(xyspeed);
                        pnpWrite(GCode);
                        
                        //drop nozzle for retraction
                        GCode = "G1 Z" + Double.toString(retractionDropHeight) + " F" + Integer.toString(zspeed);
                        pnpWrite(GCode);
                        
                        //retract tape, move it back, remove slop
                        pnpWrite("G91"); //temporarily switch to relative coordinates
                        GCode = "G1 X" + Double.toString(retractionX) + " F" + Integer.toString(retractionSpeed);
                        pnpWrite(GCode);
                        pnpWrite("G4 P200");
                        GCode = "G1 X" + Double.toString(backRetractionX) + " F" + Integer.toString(retractionSpeed);
                        pnpWrite(GCode);
                        pnpWrite("G4 P100");
                        GCode = "G1 X" + Double.toString(slopRetractionX) + " F" + Integer.toString(retractionSpeed);
                        pnpWrite(GCode);
                        pnpWrite("G90"); //return to absolute coordinates
                        
                        //lift nozzle
                        drop = reelHeight[pickUp] + 1.5;
                        GCode = "G1 Z" + Double.toString(drop) + " F" + Integer.toString(zspeed);
                        pnpWrite(GCode);
                    }

                    //move over component on tape
                    pnpWrite("G91"); //temporarily switch to relative coordinates
                    GCode = "G1 X" + Double.toString(reelPositX[pickUp]) +  " Y" + Double.toString(reelPositY[pickUp])  + " F" + Integer.toString(xyspeed);
                    pnpWrite(GCode);
                    pnpWrite("G90"); //return to absolute coordinates
                    
                    //drop nozzle on component to pick up
                    drop = reelHeight[pickUp];
                    GCode = "G1 Z" + Double.toString(drop) + " F" + Integer.toString(zspeed);
                    pnpWrite(GCode);                
                }
                
                //if tray component is received, read file to see what position to go to
                if(feedVerify == 'T'){
                    
                    //read tray position file to pickup correct component
                    int position = readTrayFile(pickUp);
                    
                    //if out of components in tray reload tray
                    if(trayCount[pickUp] < position){
                        
                        //reload tray script
                        
                    }
                    
                    //move over component on tray
                    GCode = "G1 X" + Double.toString(trayX[pickUp][position]-homeX) + " Y" + Double.toString(trayY[pickUp][position]-homeY) + " F" + Integer.toString(xyspeed);
                    pnpWrite(GCode);
                    
                    //drop nozzle on component to pick up
                    GCode = "G1 Z" + Double.toString(trayZ[pickUp]) + " F" + Integer.toString(zspeed);
                    pnpWrite(GCode);  
                }

                //turn on suction to pick up component
                if(invertNozzle == true){
                    pnpWrite("M107");
                }else{
                    pnpWrite("M106");
                }

                //dwell to ensure suction
                GCode = "G4 P" + Integer.toString(suctionDwell);
                pnpWrite(GCode);

                //lift nozzle to standard retraction
                GCode = "G1 Z" + Double.toString(standardSafetyHeight) + " F" + Integer.toString(zspeed);
                pnpWrite(GCode);
                
                //move to outer position of guides to prevent scraping on tapes
                if(feedVerify == 'R'){
                    GCode = "G1 X" + Double.toString(reelTapeX[pickUp] - 30.00) + " Y" + Double.toString(reelTapeY[pickUp]) + " F" + Integer.toString(xyspeed);
                    pnpWrite(GCode);
                }
                
                //wait for a moment to clear buffer then rotate component
                pnpWrite("G4 P1");
                GCode = "M280 P0 S" + rotationDeg;
                pnpWrite(GCode);

                //move to xposit and yposit
                GCode = "G1 X" + xposit + " Y" + yposit + " F" + Integer.toString(xyspeed);
                pnpWrite(GCode);

                //drop nozzle to board height + component height
                drop = boardHeight;
                GCode = "G1 Z" + Double.toString(drop) + " F" + Integer.toString(zspeed);
                pnpWrite(GCode);

                //turn off suction
                if(invertNozzle == true){
                    pnpWrite("M106");
                }else{
                    pnpWrite("M107");
                }

                //dwell to ensure release of component
                GCode = "G4 P" + Integer.toString(releaseDwell);
                pnpWrite(GCode);
            }
            
            //close scanner to prevent overflow?
            incomingToken.close();
        }
        
        //close all opened files
        backUpFile.close();
        
        //Return to normal state
        System.out.println("\nFile Execution Complete. Returning to Home");
        pnpWrite("G1 Z8 F100");     //raise z axis
        pnpWrite("G4 P1");
        pnpWrite("M280 P0 S0");     //put nozzle back at 0 degrees
        pnpWrite("G1 X0 Y0 F2000"); //go to zero on x and y axis
        pnpWrite("G1 Z0 F100");     //go to zero on z axis
        pnpWrite("M107");           //turn off valve
        pnpWrite("M18");            //disable steppers
    }
    
    public static void runFileTest(Scanner console) throws FileNotFoundException, SerialPortException, IOException{
        
        //ensure at origin
        pnpWrite("G1 X0 Y0 Z0 F1000");
        
        //search directory and list names of files
        File folder = new File(fileDirectory);
        File[] listOfFiles = folder.listFiles();
        System.out.println("\nContents of Directory\n");
        for(int i = 0; i < listOfFiles.length; i++) {
            if (listOfFiles[i].isFile()) {
                System.out.println("File:    " + listOfFiles[i].getName());
            } else if (listOfFiles[i].isDirectory()) {
                System.out.println("Folder:  " + listOfFiles[i].getName());
            }
        }
        
        //ask which file to run
        System.out.print("Which File to run? include extension: ");
        String fileName = console.next();
        String fullpath = fileDirectory + "/" + fileName;
        File GCodeFile = new File(fullpath);
        
        //Check if file exists
        while(!GCodeFile.exists()){
            System.out.println("File does not exist!!! ");
            System.out.print("Enter file name with extension: ");
            fileName = console.next();
            GCodeFile = new File(fileName);
        }
        System.out.println("File " + fileName + " opened.");
        
        //ask to commence execution
        System.out.print("Commence Execution? (yes or no) ");
        fileName = console.next();
        if(fileName.equals("no"))
            return;
        
        //setup common variables
        String discard =    "0";
        String x_absolute = "0";
        String y_absolute = "0";
        double xabs = 0.00;
        double yabs = 0.00; 
        String GCode = "0";
        double drop = 0.00;
        boolean setAbsolute = false;
        
        //execute file
        Scanner gcodeScanner = new Scanner(GCodeFile);
        while(gcodeScanner.hasNextLine()){
         
            //reset the set absolute attribute
            setAbsolute = false;
            
            //Consume line, copy it to backup file in case of mid pnp failure
            String incomingString = gcodeScanner.nextLine();
            Scanner incomingToken = new Scanner(incomingString);
            
            //check if new relative position expected
            char checkChar = incomingString.charAt(0);
            if(checkChar == '#'){
                setAbsolute = true;
            }
            
            //set absolute coordinates if token received to do so (received #)
            if(setAbsolute == true){
                
                //expect three tokens
                discard =   incomingToken.next();  //the # symbol needs to be consumed by scanner
                x_absolute = incomingToken.next();
                y_absolute = incomingToken.next();
                
                //define new absolute coordinates
                xabs = Double.parseDouble(x_absolute);
                yabs = Double.parseDouble(y_absolute);
                
                //tell user new absolute coordinates set
                System.out.println("\nNew Board Coordinates Set");
                System.out.println("X Posit: " + xabs);
                System.out.println("Y Posit: " + yabs + "\n");

                //lift head and go to new board coordinates
                GCode = "G1 Z" + Double.toString(standardSafetyHeight) + " F" + Integer.toString(zspeed);
                pnpWrite(GCode);
                GCode = "G1 X" + Double.toString(xabs) + " Y" + Double.toString(yabs) + " F" + Integer.toString(xyspeed);
                pnpWrite(GCode);
                
                //lower to board height
                drop = boardHeight;
                GCode = "G1 Z" + Double.toString(drop) + " F" + Integer.toString(zspeed);
                pnpWrite(GCode);
                
                //ask for permission to continue
                System.out.println("Move head to first component");
                System.in.read();
                
            }
            
            //if no new absolute coordinate, execute the pick and place of components listed
            else {
                //5 tokens are expected, put them in variables for manipulation
                String feedType = incomingToken.next();    //feed type either "R" or "T" for reel or tray
                String component = incomingToken.next();   //component number of feedType
                String xposit = incomingToken.next();      //x position to place component
                String yposit = incomingToken.next();      //y position to place component
                String rotationDeg = incomingToken.next(); //rotation of component
                
                //turn tokens into other variables as necessary
                char feedVerify = feedType.charAt(0);         
                int pickUp = Integer.parseInt(component) - 1;
                double parseX = Double.parseDouble(xposit) + xabs; //determine absolute coordinate
                double parseY = Double.parseDouble(yposit) + yabs; //determine absolute coordinate
                xposit = Double.toString(parseX); //absolute x coordinate of component to move to
                yposit = Double.toString(parseY); //absolute y coordinate of component to move to

                //lift nozzle to above board
                drop = boardHeight + 3.00;
                GCode = "G1 Z" + Double.toString(drop) + " F" + Integer.toString(zspeed);
                pnpWrite(GCode);

                //move to xposit and yposit
                GCode = "G1 X" + xposit + " Y" + yposit + " F" + Integer.toString(xyspeed);
                pnpWrite(GCode);
                
                //rotate code
                //rotate code
                //rotate code

                //drop nozzle to board height
                drop = boardHeight;
                GCode = "G1 Z" + Double.toString(drop) + " F" + Integer.toString(zspeed);
                pnpWrite(GCode);
                
                //ask for permission to continue
                System.out.println("Move head to next position");
                System.in.read();
            }
            
            //close scanner to prevent overflow?
            incomingToken.close();
        }
        
        //Return to normal state
        System.out.println("\nFile Execution Complete. Returning to Home");
        pnpWrite("G1 Z3.5 F100");   //raise z axis
        pnpWrite("G1 X0 Y0 F2000"); //go to zero on x and y axis
        pnpWrite("G1 Z0 F100");     //go to zero on z axis
        pnpWrite("M107");           //turn off valve
        pnpWrite("M18");            //disable steppers
    }
    
    
    //Sleep Until Computer receives an "okay" from the PNP
    public static void pnpWrite(String GCode) throws SerialPortException{
        System.out.println("Sent: " + GCode);  //uncomment to see what is being sent to machine
        GCode = GCode + "\n\r"; //add the required new line and carriage return
        inputStr = "busy";
        serialPort.writeString(GCode); //write to the port
        //Stay in loop until okay from machine is received
        while(true){
            try{
                Thread.sleep(5);
                if(inputStr.equals("ok")){
                    break;
                }
            }
            catch(Exception ex){
                System.out.println(ex);
            }

        }
    }
    
    //Serial Port Listener
    static class SerialPortReader implements SerialPortEventListener {
        public void serialEvent(SerialPortEvent event) {    
            if(event.isRXCHAR()){
                try{
                    Thread.sleep(20); //delay a little bit to get everything into the buffer
                }
                catch(Exception ex){
                    System.out.println(ex);
                }
                //Copy the input into decision variable if expected input recieved
                try {
                    String tempVal = serialPort.readString();
                    //System.out.print(tempVal);  //UNCOMMENT TO SEE WHAT IT IS BEING received
                    //Received okay status from command
                    if(tempVal.charAt(0) == 'o' && tempVal.charAt(1) == 'k'){
                        inputStr = "ok";
                    }
                }
                catch (SerialPortException ex) {
                    System.out.println(ex);
                }
            }
         }
     }
    
    public static int readTrayFile(int pickUp) throws FileNotFoundException, IOException{
        
        int lineToCheck = 0;
        int returnNumber = 0;
        boolean trayExists = false;

        //open tray array file and create temp file
        FileReader trayFile = new FileReader("config/trayArray.txt");
        BufferedReader buffTrayFile = new BufferedReader(trayFile);
        Scanner scanTrayFile = new Scanner(buffTrayFile);
        PrintStream tempFile = new PrintStream("config/temp.txt");

        //scan tray file to copy and modify if tray component is used
        while(scanTrayFile.hasNextLine()){

            String lineFromTrayFile = scanTrayFile.nextLine();
            if(lineToCheck == pickUp){
                int trayNumber = Integer.parseInt(lineFromTrayFile);
                returnNumber = trayNumber;
                trayNumber++;
                lineFromTrayFile = Integer.toString(trayNumber);
                trayExists = true;     
            }
            tempFile.println(lineFromTrayFile);
            lineToCheck++;
        }
        //delete old file, create new, delete temp
        Path source = Paths.get("config/temp.txt");
        Path destination = Paths.get("config/trayArray.txt");
        try {
            Files.delete(destination);
            Files.copy(source, destination);
            Files.delete(source);
        } catch (IOException e) {
                e.printStackTrace();
        }

        //close all files to prevent unforseen consequences
        tempFile.close();
        buffTrayFile.close();
        trayFile.close();
        
        //if tray doesn't exist tell user
        if(trayExists == false){
            System.out.println("Tray Doesn't Exist!!!");
            System.in.read();
        }
        return returnNumber;
    }
    
//end program
}